Package: UniCensor
Type: Package
Title: Reproducible Random Samples Under Univariate Censoring Schemes
Version: 0.1.0
Authors@R: person("Shikhar", "Tyagi",
                  email = "shikhar1093tyagi@gmail.com",
                  role  = c("aut", "cre"),
                  comment = c(ORCID = "0000-0003-1606-0844"))
Description: Generates reproducible random samples from any user-specified
    univariate distribution under a comprehensive suite of censoring and
    truncation schemes.
    Users supply the probability density function (PDF),
    cumulative distribution function (CDF), survival function, support bounds,
    and parameters; the same seed and inputs yield identical samples across
    sessions.  Supported schemes include right and left truncation, random,
    right, left, interval, and middle censoring, block random censoring,
    balanced joint progressive Type-II (BJPT-II), progressive first failure,
    joint Type-I, Type-I, Type-II, progressive Type-II, Type-II progressively
    hybrid, joint Type-II, hybrid, hybrid Type-I, doubly Type-II, Type-I
    hybrid, and hybrid Type-II censoring.  Diagnostic histogram, dot plot, and
    autocorrelation plots are provided for each scheme to verify distributional
    behaviour.
    Methods are described in
    Nagar, Kumar, and Krishna (2026) <doi:10.59467/IJASS.2026.22.1>,
    Goel, Kumar, and Krishna (2026, "Estimation in power Lindley distributions
    using balanced joint progressively Type-II censored data"),
    Wu and Kus (2009) <doi:10.1016/j.csda.2009.03.010>,
    Goel and Krishna (2026) <doi:10.1007/s13198-026-03208-w>,
    Balakrishnan and Aggarwala (2000, ISBN:978-1-4612-1334-5),
    Mondal and Kundu (2020) <doi:10.1080/03610926.2018.1554128>,
    Ding and Gui (2023) <doi:10.3390/math11092003>,
    Prajapati, Mitra, and Kundu (2019) <doi:10.1007/s13571-018-0167-0>,
    Yadav, Jaiswal, and Yadav (2026) <doi:10.1007/s11135-026-02647-8>,
    Iyer, Jammalamadaka, and Kundu (2008) <doi:10.1016/j.jspi.2007.03.062>,
    Banerjee and Kundu (2008) <doi:10.1109/TR.2008.916890>,
    and Kundu and Joarder (2006) <doi:10.1016/j.csda.2005.05.002>.
License: GPL-3
Depends: R (>= 4.0.0)
Imports: graphics, stats
Suggests: testthat (>= 3.0.0)
Config/testthat/edition: 3
Encoding: UTF-8
Language: en-US
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-07-08 23:26:00 UTC; shikhar tyagi
Author: Shikhar Tyagi [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-1606-0844>)
Maintainer: Shikhar Tyagi <shikhar1093tyagi@gmail.com>
Repository: CRAN
Date/Publication: 2026-07-17 13:40:08 UTC

# Generated by roxygen2: do not edit by hand

S3method(as.data.frame,censored_sample)
S3method(as.vector,censored_sample)
S3method(print,censored_sample)
S3method(print,dist_spec)
export(dist_spec)
export(plot_censor_acf)
export(plot_censor_diagnostics)
export(plot_censor_dot)
export(plot_censor_hist)
export(r_bjpt2_censor)
export(r_block_random_censor)
export(r_complete)
export(r_doubly_type2_censor)
export(r_hybrid_censor)
export(r_hybrid_type1_censor)
export(r_hybrid_type2_censor)
export(r_interval_censor)
export(r_joint_type1_censor)
export(r_joint_type2_censor)
export(r_left_censor)
export(r_left_truncation)
export(r_middle_censor)
export(r_progressive_first_failure)
export(r_progressive_hybrid_type2_censor)
export(r_progressive_type2_censor)
export(r_random_censor)
export(r_right_censor)
export(r_right_truncation)
export(r_type1_censor)
export(r_type1_hybrid_censor)
export(r_type2_censor)
importFrom(graphics,abline)
importFrom(graphics,hist)
importFrom(graphics,legend)
importFrom(graphics,lines)
importFrom(graphics,par)
importFrom(graphics,plot)
importFrom(graphics,points)
importFrom(stats,acf)
importFrom(stats,integrate)
importFrom(stats,qnorm)
importFrom(stats,quantile)
importFrom(stats,rexp)
importFrom(stats,runif)
importFrom(stats,uniroot)

#' Define a univariate distribution for simulation
#'
#' Creates a reusable distribution specification passed to all
#' \pkg{UniCensor} sampling functions.  At least one of \code{q},
#' \code{surv}, \code{p}, or \code{d} must be supplied for random
#' number generation via inverse transform sampling.
#'
#' @param d function; PDF \eqn{f(x, \theta)}.
#' @param p function; CDF \eqn{F(x, \theta)}.
#' @param q function; quantile function \eqn{F^{-1}(u, \theta)}.
#' @param surv function; survival function \eqn{S(x, \theta)}.
#' @param lower numeric; lower support bound (default \code{0}).
#' @param upper numeric; upper support bound (default \code{Inf}).
#' @param param named list of distribution parameters (default empty).
#' @param name character label for plots (default \code{"custom"}).
#'
#' @return An object of class \code{"dist_spec"}, which is a list with
#'   elements \code{d}, \code{p}, \code{q}, \code{surv} (user-supplied
#'   functions or \code{NULL}), \code{lower} and \code{upper} (support
#'   bounds), \code{param} (named list of parameters), and \code{name}
#'   (character label).
#' @export
#'
#' @examples
#' # Exponential(rate = 1)
#' exp_dist <- dist_spec(
#'   d    = function(x, p) p$rate * exp(-p$rate * x),
#'   p    = function(x, p) 1 - exp(-p$rate * x),
#'   q    = function(u, p) -log(1 - u) / p$rate,
#'   surv = function(x, p) exp(-p$rate * x),
#'   lower = 0, upper = Inf,
#'   param = list(rate = 1),
#'   name  = "Exp(1)"
#' )
dist_spec <- function(d = NULL, p = NULL, q = NULL, surv = NULL,
                      lower = 0, upper = Inf, param = list(),
                      name = "custom") {
  if (is.null(q) && is.null(surv) && is.null(p) && is.null(d)) {
    stop("Provide at least one of q, surv, p, or d.", call. = FALSE)
  }
  structure(
    list(
      d     = d,
      p     = p,
      q     = q,
      surv  = surv,
      lower = lower,
      upper = upper,
      param = param,
      name  = name
    ),
    class = "dist_spec"
  )
}

#' @title Print method for dist_spec
#' @param x a \code{dist_spec} object.
#' @param ... ignored.
#' @return Invisibly, the input \code{dist_spec} object \code{x}.  Called
#'   for its side effect of printing a human-readable summary of the
#'   distribution name, support, parameters, and available functions.
#' @export
print.dist_spec <- function(x, ...) {
  cat("Distribution:", x$name, "\n")
  cat("Support: [", x$lower, ", ", x$upper, "]\n", sep = "")
  cat("Parameters:", paste(names(x$param), x$param, sep = " = ",
                           collapse = ", "), "\n")
  fn <- c(d = !is.null(x$d), p = !is.null(x$p),
          q = !is.null(x$q), surv = !is.null(x$surv))
  cat("Functions:", paste(names(fn)[fn], collapse = ", "), "\n")
  invisible(x)
}

#' @title Print method for censored_sample
#' @param x a \code{censored_sample} object.
#' @param ... ignored.
#' @return Invisibly, the input \code{censored_sample} object \code{x}.
#'   Called for its side effect of printing the censoring scheme, distribution
#'   name, sample size, seed, and a frequency table of status codes.
#' @export
print.censored_sample <- function(x, ...) {
  cat("Censoring scheme:", x$scheme, "\n")
  cat("Distribution:", x$dist$name, "  n =", x$n)
  if (!is.null(x$seed)) cat("  seed =", x$seed)
  cat("\n")
  cat("Status counts:\n")
  print(table(x$status, useNA = "ifany"))
  invisible(x)
}

#' @title Extract observed values from a censored sample
#' @param x a \code{censored_sample} object.
#' @param ... ignored.
#' @return Numeric vector of observed values (failures and censoring times).
#' @export
as.vector.censored_sample <- function(x, ...) x$x

#' @title Coerce censored sample to data frame
#' @param x a \code{censored_sample} object.
#' @param row.names ignored.
#' @param optional ignored.
#' @param ... ignored.
#' @return A data frame with columns \code{x} (observed values),
#'   \code{status} (integer status code: 0 = right-censored, 1 = event,
#'   2 = left-censored, 3 = interval-censored, 4 = middle-censored),
#'   \code{t_true} (latent true failure times), and \code{time2} (upper
#'   interval bound for interval- or middle-censored observations,
#'   \code{NA} otherwise).
#' @export
as.data.frame.censored_sample <- function(x, row.names = NULL,
                                          optional = FALSE, ...) {
  data.frame(
    x      = x$x,
    status = x$status,
    t_true = x$t_true,
    time2  = x$time2,
    stringsAsFactors = FALSE
  )
}

# Diagnostic plots --------------------------------------------------------------

#' Histogram of observed values with true density overlay
#'
#' For censored samples, uses latent \code{t_true} values to overlay the
#' target density when available; observed \code{x} values are histogrammed
#' with a note on censoring status.
#'
#' @param samp a \code{censored_sample} object.
#' @param dist optional \code{dist_spec}; defaults to \code{samp$dist}.
#' @param use_latent logical; plot histogram of latent failures
#'   (\code{t_true}) instead of observed \code{x} (default \code{TRUE} for
#'   verifying the underlying distribution).
#' @param nbreaks number of histogram breaks (default \code{30}).
#' @param ... graphical parameters passed to \code{\link[graphics]{hist}}.
#' @return Invisibly, an object of class \code{"histogram"} as returned by
#'   \code{\link[graphics]{hist}}, containing break points, counts, and
#'   density values.  Called primarily for its side effect of producing a
#'   histogram plot with an optional PDF overlay.
#' @export
plot_censor_hist <- function(samp, dist = NULL, use_latent = TRUE,
                             nbreaks = 30L, ...) {
  if (!inherits(samp, "censored_sample")) {
    stop("'samp' must be a censored_sample object.", call. = FALSE)
  }
  dist <- dist %||% samp$dist
  vals <- if (use_latent) samp$t_true else samp$x
  ttl  <- paste0(samp$scheme, " - ", dist$name,
                 if (use_latent) " (latent)" else " (observed)")
  h <- graphics::hist(vals, breaks = nbreaks, freq = FALSE,
                      main = ttl, xlab = "Value", col = "steelblue",
                      border = "white", ...)
  if (!is.null(dist$d)) {
    xr <- range(vals, na.rm = TRUE)
    grid <- seq(xr[1], xr[2], length.out = 200)
    graphics::lines(grid, vapply(grid, function(t) dist$d(t, dist$param),
                                 numeric(1)), col = "darkred", lwd = 2)
    graphics::legend("topright", legend = c("Sample", "PDF"),
                     fill = c("steelblue", NA), border = c("white", NA),
                     lty = c(NA, 1), col = c(NA, "darkred"), lwd = c(NA, 2),
                     bty = "n")
  }
  invisible(h)
}

#' Dot plot (index plot) of sample values
#'
#' @param samp a \code{censored_sample} object.
#' @param use_latent logical; plot latent (\code{t_true}) or observed
#'   (\code{x}) values.
#' @param pch plotting character; censored points use \code{pch = 1},
#'   events use \code{pch = 16}.
#' @param ... passed to \code{\link[graphics]{plot}}.
#' @return Invisibly, a list with elements \code{index} (integer vector of
#'   observation indices) and \code{values} (numeric vector of plotted
#'   values).  Called primarily for its side effect of producing a dot
#'   (index) plot distinguishing events from censored observations.
#' @export
plot_censor_dot <- function(samp, use_latent = TRUE, pch = NULL, ...) {
  if (!inherits(samp, "censored_sample")) {
    stop("'samp' must be a censored_sample object.", call. = FALSE)
  }
  vals <- if (use_latent) samp$t_true else samp$x
  idx  <- seq_along(vals)
  ttl  <- paste0(samp$scheme, " dot plot - ", samp$dist$name)
  graphics::plot(idx, vals, type = "n", main = ttl,
                 xlab = "Index", ylab = "Value", ...)
  ev  <- samp$status == 1L
  cen <- !ev
  if (is.null(pch)) {
    graphics::points(idx[ev], vals[ev], pch = 16, col = "darkgreen")
    if (any(cen)) {
      graphics::points(idx[cen], vals[cen], pch = 1, col = "firebrick")
    }
    graphics::legend("topright",
                     legend = c("Event", "Censored"),
                     pch = c(16, 1), col = c("darkgreen", "firebrick"),
                     bty = "n")
  } else {
    graphics::points(idx, vals, pch = pch, ...)
  }
  invisible(list(index = idx, values = vals))
}

#' Autocorrelation function plot
#'
#' Plots the ACF of latent failure times to assess independence of the
#' underlying RNG stream (should resemble white noise for valid simulation).
#'
#' @param samp a \code{censored_sample} object.
#' @param use_latent logical; compute ACF on \code{t_true} (default) or
#'   observed \code{x}.
#' @param max_lag maximum lag (default \code{min(20, n/4)}).
#' @param ... passed to \code{\link[graphics]{plot}}.
#' @return Invisibly, an object of class \code{"acf"} as returned by
#'   \code{\link[stats]{acf}}, containing the autocorrelation values at each
#'   lag.  Called primarily for its side effect of producing an ACF plot
#'   with 95\% confidence bands.
#' @export
plot_censor_acf <- function(samp, use_latent = TRUE, max_lag = NULL, ...) {
  if (!inherits(samp, "censored_sample")) {
    stop("'samp' must be a censored_sample object.", call. = FALSE)
  }
  vals <- if (use_latent) samp$t_true else samp$x
  n    <- length(vals)
  if (is.null(max_lag)) max_lag <- min(20L, max(1L, n %/% 4L))
  acf_obj <- stats::acf(vals, lag.max = max_lag, plot = FALSE)
  ttl <- paste0(samp$scheme, " ACF - ", samp$dist$name)
  ci  <- stats::qnorm(0.975) / sqrt(n)
  graphics::plot(acf_obj$lag[-1], acf_obj$acf[-1], type = "h",
                 ylim = c(-1, 1), main = ttl, xlab = "Lag", ylab = "ACF",
                 ...)
  graphics::abline(h = c(-ci, ci), lty = 2, col = "gray50")
  graphics::abline(h = 0)
  invisible(acf_obj)
}

#' Combined diagnostic panel: histogram, dot plot, and ACF
#'
#' Produces a three-panel plot (histogram with PDF overlay, dot plot, and
#' ACF) for a censored sample.  Graphical parameters are saved before
#' modification and restored on exit via \code{\link{on.exit}}.
#'
#' @param samp a \code{censored_sample} object.
#' @param dist optional \code{dist_spec}.
#' @param use_latent passed to individual plot functions.
#' @param ... ignored.
#' @return Invisibly, the input \code{censored_sample} object \code{samp}.
#'   Called primarily for its side effect of producing a combined diagnostic
#'   plot panel.
#' @export
plot_censor_diagnostics <- function(samp, dist = NULL, use_latent = TRUE,
                                    ...) {
  if (!inherits(samp, "censored_sample")) {
    stop("'samp' must be a censored_sample object.", call. = FALSE)
  }
  oldpar <- graphics::par(no.readonly = TRUE)
  on.exit(graphics::par(oldpar))
  graphics::par(mfrow = c(1, 3), mar = c(4, 4, 3, 1))
  plot_censor_hist(samp, dist = dist, use_latent = use_latent)
  plot_censor_dot(samp, use_latent = use_latent)
  plot_censor_acf(samp, use_latent = use_latent)
  invisible(samp)
}

#' @keywords internal
`%||%` <- function(x, y) if (is.null(x)) y else x

#' @title UniCensor: Reproducible Univariate Samples Under Censoring Schemes
#'
#' @description
#' The \pkg{UniCensor} package simulates censored or truncated random samples
#' from any user-defined univariate distribution.  Supply the PDF, CDF,
#' survival function, support bounds, parameters, sample size, and a seed;
#' identical inputs always produce identical output.
#'
#' @section Distribution specification:
#' Use \code{\link{dist_spec}} to define a distribution object, then pass it
#' to any scheme-specific generator (e.g. \code{\link{r_type1_censor}}).
#'
#' @section Censoring schemes (one function each):
#' \describe{
#'   \item{\code{\link{r_complete}}}{Uncensored (complete) data.}
#'   \item{\code{\link{r_right_truncation}}}{Right truncation.}
#'   \item{\code{\link{r_left_truncation}}}{Left truncation.}
#'   \item{\code{\link{r_random_censor}}}{Random censoring.}
#'   \item{\code{\link{r_right_censor}}}{Right censoring.}
#'   \item{\code{\link{r_left_censor}}}{Left censoring.}
#'   \item{\code{\link{r_interval_censor}}}{Interval censoring.}
#'   \item{\code{\link{r_middle_censor}}}{Middle censoring.}
#'   \item{\code{\link{r_block_random_censor}}}{Block random censoring.}
#'   \item{\code{\link{r_bjpt2_censor}}}{Balanced joint progressive Type-II.}
#'   \item{\code{\link{r_progressive_first_failure}}}{Progressive first failure.}
#'   \item{\code{\link{r_joint_type1_censor}}}{Joint Type-I censoring.}
#'   \item{\code{\link{r_type1_censor}}}{Type-I (time-terminated).}
#'   \item{\code{\link{r_type2_censor}}}{Type-II (failure-terminated).}
#'   \item{\code{\link{r_progressive_type2_censor}}}{Progressive Type-II.}
#'   \item{\code{\link{r_progressive_hybrid_type2_censor}}}{Type-II progressively hybrid.}
#'   \item{\code{\link{r_joint_type2_censor}}}{Joint Type-II censoring.}
#'   \item{\code{\link{r_hybrid_censor}}}{Hybrid censoring.}
#'   \item{\code{\link{r_hybrid_type1_censor}}}{Hybrid Type-I censoring.}
#'   \item{\code{\link{r_doubly_type2_censor}}}{Doubly Type-II censoring.}
#'   \item{\code{\link{r_type1_hybrid_censor}}}{Type-I hybrid censoring.}
#'   \item{\code{\link{r_hybrid_type2_censor}}}{Hybrid Type-II censoring.}
#' }
#'
#' @section Diagnostic plots:
#' \code{\link{plot_censor_hist}}, \code{\link{plot_censor_dot}},
#' \code{\link{plot_censor_acf}}, and \code{\link{plot_censor_diagnostics}}
#' (all three panels).
#'
#' @author Shikhar Tyagi \email{shikhar1093tyagi@@gmail.com}
#' @importFrom stats acf integrate qnorm quantile rexp runif uniroot
#' @importFrom graphics abline hist legend lines par plot points
#' @keywords internal
"_PACKAGE"

# Complete (uncensored) sample ------------------------------------------------

#' Generate a complete (uncensored) random sample
#'
#' @param n sample size.
#' @param dist a \code{\link{dist_spec}} object.
#' @param seed optional integer seed for reproducibility.
#' @return A \code{censored_sample} object (S3 class) containing:
#'   \code{scheme} (character), \code{dist} (the \code{dist_spec}),
#'   \code{n} (integer sample size), \code{seed}, \code{t_true} (numeric
#'   vector of latent failure times), \code{x} (observed/censored values),
#'   \code{status} (integer: 1 = event for all observations in a complete
#'   sample), and \code{time2} (\code{NA} for all).
#' @export
#' @examples
#' exp_dist <- dist_spec(
#'   q = function(u, p) -log(1 - u) / p$rate,
#'   surv = function(x, p) exp(-p$rate * x),
#'   param = list(rate = 1), name = "Exp(1)"
#' )
#' samp <- r_complete(50, exp_dist, seed = 42)
#' table(samp$status)
r_complete <- function(n, dist, seed = NULL) {
  n <- .check_n(n)
  .validate_dist(dist)
  .apply_seed(seed)
  t_true <- .rng_stream(n, dist)
  .as_censored_sample("complete", dist, n, seed, t_true, t_true,
                      rep(1L, n))
}

# Truncation -------------------------------------------------------------------

#' Right-truncated random sample
#'
#' Observations are drawn from \eqn{X \mid X \le \code{right}}.
#'
#' @param n sample size.
#' @param dist a \code{\link{dist_spec}} object.
#' @param right upper truncation point.
#' @param seed optional integer seed.
#' @return A \code{censored_sample} object (S3 class) containing:
#'   \code{scheme} (character), \code{dist} (the \code{dist_spec}),
#'   \code{n} (integer sample size), \code{seed}, \code{t_true} (numeric
#'   vector of latent failure times), \code{x} (observed/censored values),
#'   \code{status} (integer: 0 = right-censored, 1 = event, 2 = left-censored,
#'   3 = interval-censored, 4 = middle-censored), and \code{time2}
#'   (upper interval bound where applicable, \code{NA} otherwise).
#' @export
#' @examples
#' d <- dist_spec(p = function(x, p) 1 - exp(-p$rate * x),
#'                q = function(u, p) -log(1 - u) / p$rate,
#'                param = list(rate = 1), name = "Exp(1)")
#' r_right_truncation(20, d, right = 3, seed = 1)
r_right_truncation <- function(n, dist, right, seed = NULL) {
  n <- .check_n(n)
  .validate_dist(dist)
  .apply_seed(seed)
  t_true <- .rng_truncated(n, dist, left = dist$lower, right = right)
  .as_censored_sample("right_truncation", dist, n, seed, t_true, t_true,
                      rep(1L, n), extra = list(right = right))
}

#' Left-truncated random sample
#'
#' Observations are drawn from \eqn{X \mid X \ge \code{left}}.
#'
#' @param n sample size.
#' @param dist a \code{\link{dist_spec}} object.
#' @param left lower truncation point.
#' @param seed optional integer seed.
#' @return A \code{censored_sample} object (S3 class) containing:
#'   \code{scheme} (character), \code{dist} (the \code{dist_spec}),
#'   \code{n} (integer sample size), \code{seed}, \code{t_true} (numeric
#'   vector of latent failure times), \code{x} (observed/censored values),
#'   \code{status} (integer: 0 = right-censored, 1 = event, 2 = left-censored,
#'   3 = interval-censored, 4 = middle-censored), and \code{time2}
#'   (upper interval bound where applicable, \code{NA} otherwise).
#' @export
#' @examples
#' d <- dist_spec(p = function(x, p) 1 - exp(-p$rate * x),
#'                q = function(u, p) -log(1 - u) / p$rate,
#'                param = list(rate = 1), name = "Exp(1)")
#' r_left_truncation(20, d, left = 0.5, seed = 1)
r_left_truncation <- function(n, dist, left, seed = NULL) {
  n <- .check_n(n)
  .validate_dist(dist)
  .apply_seed(seed)
  upper <- if (is.finite(dist$upper)) dist$upper else left + 1e4
  t_true <- .rng_truncated(n, dist, left = left, right = upper)
  .as_censored_sample("left_truncation", dist, n, seed, t_true, t_true,
                      rep(1L, n), extra = list(left = left))
}

# Simple censoring -------------------------------------------------------------

#' Random censoring
#'
#' Each subject is observed until \eqn{\min(X, C)} where \eqn{C} follows an
#' exponential distribution with rate \code{cen_rate}.
#'
#' @param n sample size.
#' @param dist a \code{\link{dist_spec}} object.
#' @param cen_rate exponential censoring rate (default \code{0.2}).
#' @param seed optional integer seed.
#' @return A \code{censored_sample} object (S3 class) containing:
#'   \code{scheme} (character), \code{dist} (the \code{dist_spec}),
#'   \code{n} (integer sample size), \code{seed}, \code{t_true} (numeric
#'   vector of latent failure times), \code{x} (observed/censored values),
#'   \code{status} (integer: 0 = right-censored, 1 = event, 2 = left-censored,
#'   3 = interval-censored, 4 = middle-censored), and \code{time2}
#'   (upper interval bound where applicable, \code{NA} otherwise).
#' @export
#' @examples
#' d <- dist_spec(q = function(u, p) -log(1 - u) / p$rate,
#'                param = list(rate = 1), name = "Exp(1)")
#' r_random_censor(30, d, seed = 1)
r_random_censor <- function(n, dist, cen_rate = 0.2, seed = NULL) {
  n <- .check_n(n)
  .validate_dist(dist)
  .apply_seed(seed)
  t_true <- .rng_stream(n, dist)
  c_time <- stats::rexp(n, rate = cen_rate)
  cens   <- t_true > c_time
  x      <- ifelse(cens, c_time, t_true)
  status <- ifelse(cens, 0L, 1L)
  .as_censored_sample("random", dist, n, seed, t_true, x, status,
                      extra = list(cen_rate = cen_rate))
}

#' Right censoring at a fixed time
#'
#' @param n sample size.
#' @param dist a \code{\link{dist_spec}} object.
#' @param cen_time fixed right-censoring time.
#' @param seed optional integer seed.
#' @return A \code{censored_sample} object (S3 class) containing:
#'   \code{scheme} (character), \code{dist} (the \code{dist_spec}),
#'   \code{n} (integer sample size), \code{seed}, \code{t_true} (numeric
#'   vector of latent failure times), \code{x} (observed/censored values),
#'   \code{status} (integer: 0 = right-censored, 1 = event, 2 = left-censored,
#'   3 = interval-censored, 4 = middle-censored), and \code{time2}
#'   (upper interval bound where applicable, \code{NA} otherwise).
#' @export
#' @examples
#' d <- dist_spec(q = function(u, p) -log(1 - u) / p$rate,
#'                param = list(rate = 1), name = "Exp(1)")
#' r_right_censor(30, d, cen_time = 2, seed = 1)
r_right_censor <- function(n, dist, cen_time, seed = NULL) {
  n <- .check_n(n)
  .validate_dist(dist)
  .apply_seed(seed)
  t_true <- .rng_stream(n, dist)
  cens   <- t_true > cen_time
  x      <- pmin(t_true, cen_time)
  status <- ifelse(cens, 0L, 1L)
  .as_censored_sample("right", dist, n, seed, t_true, x, status,
                      extra = list(cen_time = cen_time))
}

#' Left censoring at a fixed threshold
#'
#' @param n sample size.
#' @param dist a \code{\link{dist_spec}} object.
#' @param threshold left-censoring threshold.
#' @param seed optional integer seed.
#' @return A \code{censored_sample} object (S3 class) containing:
#'   \code{scheme} (character), \code{dist} (the \code{dist_spec}),
#'   \code{n} (integer sample size), \code{seed}, \code{t_true} (numeric
#'   vector of latent failure times), \code{x} (observed/censored values),
#'   \code{status} (integer: 0 = right-censored, 1 = event, 2 = left-censored,
#'   3 = interval-censored, 4 = middle-censored), and \code{time2}
#'   (upper interval bound where applicable, \code{NA} otherwise).
#' @export
#' @examples
#' d <- dist_spec(q = function(u, p) -log(1 - u) / p$rate,
#'                param = list(rate = 1), name = "Exp(1)")
#' r_left_censor(30, d, threshold = 0.5, seed = 1)
r_left_censor <- function(n, dist, threshold, seed = NULL) {
  n <- .check_n(n)
  .validate_dist(dist)
  .apply_seed(seed)
  t_true <- .rng_stream(n, dist)
  cens   <- t_true < threshold
  x      <- ifelse(cens, threshold, t_true)
  status <- ifelse(cens, 2L, 1L)
  .as_censored_sample("left", dist, n, seed, t_true, x, status,
                      extra = list(threshold = threshold))
}

#' Interval censoring
#'
#' A proportion \code{prop_interval} of subjects are interval-censored into
#' windows of width \code{width}; optional random right censoring applies to
#' the remainder at rate \code{cen_rate}.
#'
#' @param n sample size.
#' @param dist a \code{\link{dist_spec}} object.
#' @param width interval window width.
#' @param prop_interval fraction interval-censored (default \code{0.3}).
#' @param cen_rate exponential rate for additional right censoring
#'   (default \code{0.2}).
#' @param seed optional integer seed.
#' @return A \code{censored_sample} object (S3 class) containing:
#'   \code{scheme} (character), \code{dist} (the \code{dist_spec}),
#'   \code{n} (integer sample size), \code{seed}, \code{t_true} (numeric
#'   vector of latent failure times), \code{x} (observed/censored values),
#'   \code{status} (integer: 0 = right-censored, 1 = event, 2 = left-censored,
#'   3 = interval-censored, 4 = middle-censored), and \code{time2}
#'   (upper interval bound where applicable, \code{NA} otherwise).
#' @export
#' @examples
#' d <- dist_spec(q = function(u, p) -log(1 - u) / p$rate,
#'                param = list(rate = 1), name = "Exp(1)")
#' r_interval_censor(30, d, width = 0.5, seed = 1)
r_interval_censor <- function(n, dist, width, prop_interval = 0.3,
                              cen_rate = 0.2, seed = NULL) {
  n <- .check_n(n)
  .validate_dist(dist)
  .apply_seed(seed)
  t_true  <- .rng_stream(n, dist)
  x       <- t_true
  time2   <- rep(NA_real_, n)
  status  <- rep(1L, n)

  n_int <- floor(prop_interval * n)
  if (n_int > 0L) {
    int_idx <- sample(n, n_int)
    t1 <- floor(t_true[int_idx] / width) * width
    t2 <- t1 + width
    x[int_idx]      <- pmax(t1, dist$lower)
    time2[int_idx]  <- t2
    status[int_idx] <- 3L
  }

  rem <- setdiff(seq_len(n), which(status == 3L))
  rc_frac <- 1 - prop_interval            # remaining subjects exposed to RC
  n_rc <- floor(rc_frac * length(rem))
  if (n_rc > 0L && length(rem) > 0L) {
    rc_idx <- sample(rem, min(n_rc, length(rem)))
    c_time <- stats::rexp(length(rc_idx), rate = cen_rate)
    x[rc_idx] <- pmin(t_true[rc_idx], c_time)
    status[rc_idx] <- ifelse(t_true[rc_idx] <= c_time, 1L, 0L)
  }

  .as_censored_sample("interval", dist, n, seed, t_true, x, status,
                      time2 = time2,
                      extra = list(width = width,
                                   prop_interval = prop_interval,
                                   cen_rate = cen_rate))
}

#' Middle censoring
#'
#' Observations that fall within the interval \code{[lower, upper]} are
#' middle-censored; observations outside this interval are observed exactly.
#'
#' @param n sample size.
#' @param dist a \code{\link{dist_spec}} object.
#' @param lower lower bound of the middle censoring interval.
#' @param upper upper bound of the middle censoring interval.
#' @param seed optional integer seed.
#' @return A \code{censored_sample} object (S3 class) containing:
#'   \code{scheme} (character), \code{dist} (the \code{dist_spec}),
#'   \code{n} (integer sample size), \code{seed}, \code{t_true} (numeric
#'   vector of latent failure times), \code{x} (observed/censored values),
#'   \code{status} (integer: 0 = right-censored, 1 = event, 2 = left-censored,
#'   3 = interval-censored, 4 = middle-censored), and \code{time2}
#'   (upper interval bound where applicable, \code{NA} otherwise).
#' @export
#' @examples
#' d <- dist_spec(q = function(u, p) -log(1 - u) / p$rate,
#'                param = list(rate = 1), name = "Exp(1)")
#' r_middle_censor(30, d, lower = 0.5, upper = 1.5, seed = 1)
r_middle_censor <- function(n, dist, lower, upper, seed = NULL) {
  n <- .check_n(n)
  .validate_dist(dist)
  if (lower >= upper) {
    stop("'lower' must be strictly less than 'upper'.", call. = FALSE)
  }
  .apply_seed(seed)
  t_true <- .rng_stream(n, dist)
  x      <- t_true
  status <- rep(1L, n)
  time2  <- rep(NA_real_, n)

  # Middle censoring: observations in [lower, upper] are censored
  middle <- t_true >= lower & t_true <= upper
  if (any(middle)) {
    # For middle-censored observations, report the interval boundaries
    # x stores the lower bound, time2 stores the upper bound
    x[middle]      <- lower
    time2[middle]  <- upper
    status[middle] <- 4L  # Status 4 = middle-censored
  }

  .as_censored_sample("middle", dist, n, seed, t_true, x, status,
                      time2 = time2,
                      extra = list(lower = lower, upper = upper))
}

#' Block random censoring
#'
#' Subjects are partitioned into blocks of size \code{block_size}; entire
#' blocks are randomly selected for right censoring at the block maximum
#' observed failure time or at an exponential censoring time, whichever is
#' smaller.
#'
#' @param n sample size.
#' @param dist a \code{\link{dist_spec}} object.
#' @param block_size block length (default \code{5}).
#' @param prop_blocks fraction of blocks censored (default \code{0.4}).
#' @param cen_rate exponential censoring rate within censored blocks.
#' @param seed optional integer seed.
#' @return A \code{censored_sample} object (S3 class) containing:
#'   \code{scheme} (character), \code{dist} (the \code{dist_spec}),
#'   \code{n} (integer sample size), \code{seed}, \code{t_true} (numeric
#'   vector of latent failure times), \code{x} (observed/censored values),
#'   \code{status} (integer: 0 = right-censored, 1 = event, 2 = left-censored,
#'   3 = interval-censored, 4 = middle-censored), and \code{time2}
#'   (upper interval bound where applicable, \code{NA} otherwise).
#' @export
#' @examples
#' d <- dist_spec(q = function(u, p) -log(1 - u) / p$rate,
#'                param = list(rate = 1), name = "Exp(1)")
#' r_block_random_censor(30, d, seed = 1)
r_block_random_censor <- function(n, dist, block_size = 5L,
                                  prop_blocks = 0.4, cen_rate = 0.2,
                                  seed = NULL) {
  n <- .check_n(n)
  .validate_dist(dist)
  .apply_seed(seed)
  t_true <- .rng_stream(n, dist)
  x      <- t_true
  status <- rep(1L, n)

  n_blocks <- ceiling(n / block_size)
  blocks   <- split(seq_len(n), rep(seq_len(n_blocks),
                                    each = block_size, length.out = n))
  n_cen_blocks <- floor(prop_blocks * n_blocks)
  if (n_cen_blocks > 0L) {
    cen_block_ids <- sample(n_blocks, n_cen_blocks)
    for (b in cen_block_ids) {
      idx <- blocks[[b]]
      c_time <- stats::rexp(1, rate = cen_rate)
      block_cen <- pmin(max(t_true[idx]), c_time)
      cens <- t_true[idx] > block_cen
      x[idx]      <- ifelse(cens, block_cen, t_true[idx])
      status[idx] <- ifelse(cens, 0L, 1L)
    }
  }

  .as_censored_sample("block_random", dist, n, seed, t_true, x, status,
                      extra = list(block_size = block_size,
                                   prop_blocks = prop_blocks,
                                   cen_rate = cen_rate))
}

# Type-I / Type-II -------------------------------------------------------------

#' Type-I (time-terminated) censoring
#'
#' All units are observed until fixed time \code{cen_time}; failures after
#' that time are right-censored.
#'
#' @param n sample size.
#' @param dist a \code{\link{dist_spec}} object.
#' @param cen_time fixed censoring time (default: 70th percentile of latent
#'   failure times from a pilot draw).
#' @param seed optional integer seed.
#' @return A \code{censored_sample} object (S3 class) containing:
#'   \code{scheme} (character), \code{dist} (the \code{dist_spec}),
#'   \code{n} (integer sample size), \code{seed}, \code{t_true} (numeric
#'   vector of latent failure times), \code{x} (observed/censored values),
#'   \code{status} (integer: 0 = right-censored, 1 = event, 2 = left-censored,
#'   3 = interval-censored, 4 = middle-censored), and \code{time2}
#'   (upper interval bound where applicable, \code{NA} otherwise).
#' @export
#' @examples
#' d <- dist_spec(q = function(u, p) -log(1 - u) / p$rate,
#'                param = list(rate = 1), name = "Exp(1)")
#' r_type1_censor(30, d, cen_time = 2, seed = 1)
r_type1_censor <- function(n, dist, cen_time = NULL, seed = NULL) {
  n <- .check_n(n)
  .validate_dist(dist)
  .apply_seed(seed)
  t_true <- .rng_stream(n, dist)
  if (is.null(cen_time)) cen_time <- .default_cen_time(t_true)
  cens   <- t_true > cen_time
  x      <- pmin(t_true, cen_time)
  status <- ifelse(cens, 0L, 1L)
  .as_censored_sample("type1", dist, n, seed, t_true, x, status,
                      extra = list(cen_time = cen_time))
}

#' Type-II (failure-terminated) censoring
#'
#' Experiment stops at the \code{r_failures}-th observed failure; remaining
#' units are right-censored at that time.
#'
#' @param n sample size.
#' @param dist a \code{\link{dist_spec}} object.
#' @param r_failures number of failures before termination (default
#'   \eqn{\lfloor 0.7n \rfloor}).
#' @param seed optional integer seed.
#' @return A \code{censored_sample} object (S3 class) containing:
#'   \code{scheme} (character), \code{dist} (the \code{dist_spec}),
#'   \code{n} (integer sample size), \code{seed}, \code{t_true} (numeric
#'   vector of latent failure times), \code{x} (observed/censored values),
#'   \code{status} (integer: 0 = right-censored, 1 = event, 2 = left-censored,
#'   3 = interval-censored, 4 = middle-censored), and \code{time2}
#'   (upper interval bound where applicable, \code{NA} otherwise).
#' @export
#' @examples
#' d <- dist_spec(q = function(u, p) -log(1 - u) / p$rate,
#'                param = list(rate = 1), name = "Exp(1)")
#' r_type2_censor(30, d, r_failures = 20, seed = 1)
r_type2_censor <- function(n, dist, r_failures = NULL, seed = NULL) {
  n <- .check_n(n)
  .validate_dist(dist)
  .apply_seed(seed)
  t_true <- .rng_stream(n, dist)
  if (is.null(r_failures)) r_failures <- .default_r_failures(n)
  stop <- .type2_stop(t_true, r_failures)
  .as_censored_sample("type2", dist, n, seed, t_true, stop$x, stop$status,
                      extra = list(r_failures = min(as.integer(r_failures), n),
                                   r_th_time = stop$r_th_time))
}

#' Progressive Type-II censoring
#'
#' At the \code{j}-th failure time, \code{scheme[j]} surviving units are
#' randomly removed from the risk set.
#'
#' @param n sample size.
#' @param dist a \code{\link{dist_spec}} object.
#' @param scheme integer vector of removals at successive failure times.
#' @param seed optional integer seed.
#' @return A \code{censored_sample} object (S3 class) containing:
#'   \code{scheme} (character), \code{dist} (the \code{dist_spec}),
#'   \code{n} (integer sample size), \code{seed}, \code{t_true} (numeric
#'   vector of latent failure times), \code{x} (observed/censored values),
#'   \code{status} (integer: 0 = right-censored, 1 = event, 2 = left-censored,
#'   3 = interval-censored, 4 = middle-censored), and \code{time2}
#'   (upper interval bound where applicable, \code{NA} otherwise).
#' @export
#' @examples
#' d <- dist_spec(q = function(u, p) -log(1 - u) / p$rate,
#'                param = list(rate = 1), name = "Exp(1)")
#' r_progressive_type2_censor(30, d, seed = 1)
r_progressive_type2_censor <- function(n, dist, scheme = NULL, seed = NULL) {
  n <- .check_n(n)
  .validate_dist(dist)
  .apply_seed(seed)
  t_true <- .rng_stream(n, dist)
  if (is.null(scheme)) scheme <- rep(1L, min(10L, max(1L, n %/% 5L)))
  pr <- .progressive_type2_removals(t_true, scheme)
  x <- t_true
  status <- ifelse(pr$removed, 0L, 1L)
  x[pr$removed] <- pr$censor_at[pr$removed]
  .as_censored_sample("progressive_type2", dist, n, seed, t_true, x, status,
                      extra = list(scheme = as.integer(scheme)))
}

#' Progressive first-failure censoring
#'
#' At each failure one surviving unit is randomly removed until only one unit
#' remains under observation.
#'
#' @param n sample size.
#' @param dist a \code{\link{dist_spec}} object.
#' @param seed optional integer seed.
#' @return A \code{censored_sample} object (S3 class) containing:
#'   \code{scheme} (character), \code{dist} (the \code{dist_spec}),
#'   \code{n} (integer sample size), \code{seed}, \code{t_true} (numeric
#'   vector of latent failure times), \code{x} (observed/censored values),
#'   \code{status} (integer: 0 = right-censored, 1 = event, 2 = left-censored,
#'   3 = interval-censored, 4 = middle-censored), and \code{time2}
#'   (upper interval bound where applicable, \code{NA} otherwise).
#' @export
#' @examples
#' d <- dist_spec(q = function(u, p) -log(1 - u) / p$rate,
#'                param = list(rate = 1), name = "Exp(1)")
#' r_progressive_first_failure(15, d, seed = 1)
r_progressive_first_failure <- function(n, dist, seed = NULL) {
  n <- .check_n(n)
  .validate_dist(dist)
  .apply_seed(seed)
  t_true  <- .rng_stream(n, dist)
  ord     <- order(t_true)
  removed <- logical(n)
  n_remove <- max(0L, n - 1L)

  censor_at <- rep(NA_real_, n)
  for (j in seq_len(n_remove)) {
    fail_pos  <- ord[j]
    t_j       <- t_true[fail_pos]
    remaining <- ord[which(!removed[ord] & ord > fail_pos)]
    if (length(remaining) == 0L) break
    pick <- sample(remaining, 1L)
    removed[pick] <- TRUE
    censor_at[pick] <- t_j
  }

  x      <- t_true
  status <- ifelse(removed, 0L, 1L)
  x[removed] <- censor_at[removed]
  .as_censored_sample("progressive_first_failure", dist, n, seed,
                      t_true, x, status)
}

# Joint and hybrid schemes -----------------------------------------------------

#' Joint Type-I censoring across multiple samples
#'
#' Generates \code{k} independent samples of size \code{n} each, all
#' censored at a common fixed time \code{cen_time}.
#'
#' @param n sample size per group.
#' @param k number of independent samples (default \code{2}).
#' @param dist a \code{\link{dist_spec}} object.
#' @param cen_time fixed censoring time.
#' @param seed optional integer seed.
#' @return A list of \code{censored_sample} objects (class
#'   \code{"censored_sample_list"}), one per group.  Each element contains
#'   \code{scheme}, \code{dist}, \code{n}, \code{seed}, \code{t_true},
#'   \code{x}, \code{status} (0 = right-censored, 1 = event), \code{time2},
#'   and extra fields \code{group} and \code{k}.
#' @export
#' @examples
#' d <- dist_spec(q = function(u, p) -log(1 - u) / p$rate,
#'                param = list(rate = 1), name = "Exp(1)")
#' r_joint_type1_censor(20, d, cen_time = 2, seed = 1)
r_joint_type1_censor <- function(n, dist, cen_time, k = 2L, seed = NULL) {
  n <- .check_n(n)
  k <- as.integer(k)
  .validate_dist(dist)
  .apply_seed(seed)
  groups <- vector("list", k)
  for (g in seq_len(k)) {
    t_true <- .rng_stream(n, dist)
    cens   <- t_true > cen_time
    x      <- pmin(t_true, cen_time)
    status <- ifelse(cens, 0L, 1L)
    groups[[g]] <- .as_censored_sample(
      "joint_type1", dist, n, seed, t_true, x, status,
      extra = list(cen_time = cen_time, group = g, k = k)
    )
  }
  class(groups) <- "censored_sample_list"
  groups
}

#' Joint Type-II censoring across multiple samples
#'
#' All \code{k} samples share the same termination rule: stop when a total of
#' \code{r_failures} events have been observed across all groups combined.
#'
#' @param n sample size per group.
#' @param k number of groups (default \code{2}).
#' @param dist a \code{\link{dist_spec}} object.
#' @param r_failures total failure count across all groups.
#' @param seed optional integer seed.
#' @return A list of \code{censored_sample} objects (class
#'   \code{"censored_sample_list"}), one per group.  Each element contains
#'   \code{scheme}, \code{dist}, \code{n}, \code{seed}, \code{t_true},
#'   \code{x}, \code{status} (0 = right-censored, 1 = event), \code{time2},
#'   and extra fields \code{group} and \code{k}.
#' @export
#' @examples
#' d <- dist_spec(q = function(u, p) -log(1 - u) / p$rate,
#'                param = list(rate = 1), name = "Exp(1)")
#' r_joint_type2_censor(20, d, r_failures = 15, seed = 1)
r_joint_type2_censor <- function(n, dist, r_failures, k = 2L, seed = NULL) {
  n <- .check_n(n)
  k <- as.integer(k)
  .validate_dist(dist)
  .apply_seed(seed)
  t_all <- matrix(.rng_stream(n * k, dist), nrow = n, ncol = k)
  flat  <- as.vector(t_all)
  ord   <- order(flat)
  r_failures <- min(as.integer(r_failures), n * k)
  r_th <- flat[ord[r_failures]]
  event <- t_all <= r_th
  groups <- vector("list", k)
  for (g in seq_len(k)) {
    t_true <- t_all[, g]
    x      <- ifelse(event[, g], t_true, r_th)
    status <- ifelse(event[, g], 1L, 0L)
    groups[[g]] <- .as_censored_sample(
      "joint_type2", dist, n, seed, t_true, x, status,
      extra = list(r_failures = r_failures, r_th_time = r_th,
                   group = g, k = k)
    )
  }
  class(groups) <- "censored_sample_list"
  groups
}

#' Balanced joint progressive Type-II (BJPT-II) censoring
#'
#' \code{k} samples of size \code{n} are progressively Type-II censored with
#' a balanced removal scheme \code{scheme} applied jointly at each failure
#' time across all groups.
#'
#' @param n sample size per group.
#' @param k number of groups (default \code{2}).
#' @param dist a \code{\link{dist_spec}} object.
#' @param scheme integer vector of total removals at each failure (split
#'   evenly across groups).
#' @param seed optional integer seed.
#' @return A list of \code{censored_sample} objects (class
#'   \code{"censored_sample_list"}), one per group.  Each element contains
#'   \code{scheme}, \code{dist}, \code{n}, \code{seed}, \code{t_true},
#'   \code{x}, \code{status} (0 = right-censored, 1 = event), \code{time2},
#'   and extra fields \code{group} and \code{k}.
#' @export
#' @examples
#' d <- dist_spec(q = function(u, p) -log(1 - u) / p$rate,
#'                param = list(rate = 1), name = "Exp(1)")
#' r_bjpt2_censor(20, d, k = 2, seed = 1)
r_bjpt2_censor <- function(n, dist, scheme = NULL, k = 2L, seed = NULL) {
  n <- .check_n(n)
  k <- as.integer(k)
  .validate_dist(dist)
  .apply_seed(seed)
  if (is.null(scheme)) scheme <- rep(1L, min(10L, max(1L, n %/% 5L)))
  scheme <- as.integer(scheme)

  t_mat <- matrix(.rng_stream(n * k, dist), nrow = n, ncol = k)
  removed <- matrix(FALSE, nrow = n, ncol = k)
  x_mat <- t_mat
  ids <- expand.grid(row = seq_len(n), col = seq_len(k))
  flat_time <- t_mat[cbind(ids$row, ids$col)]
  ord <- order(flat_time)

  m <- min(length(scheme), n * k - 1L)
  for (j in seq_len(m)) {
    pos <- ord[j]
    r <- ids$row[pos]
    c <- ids$col[pos]
    t_j <- t_mat[r, c]
    at_risk <- which(!removed & t_mat > t_j, arr.ind = TRUE)
    if (nrow(at_risk) == 0L) next
    n_remove <- min(scheme[j], nrow(at_risk))
    pick <- at_risk[sample(nrow(at_risk), n_remove), , drop = FALSE]
    removed[cbind(pick[, "row"], pick[, "col"])] <- TRUE
    x_mat[cbind(pick[, "row"], pick[, "col"])] <- t_j
  }

  groups <- vector("list", k)
  for (g in seq_len(k)) {
    t_true <- t_mat[, g]
    rem_g  <- removed[, g]
    x      <- x_mat[, g]
    status <- ifelse(rem_g, 0L, 1L)
    groups[[g]] <- .as_censored_sample(
      "bjpt2", dist, n, seed, t_true, x, status,
      extra = list(scheme = scheme, group = g, k = k)
    )
  }
  class(groups) <- "censored_sample_list"
  groups
}

#' Hybrid censoring (Type-I and Type-II constraints)
#'
#' Stops at \code{cen_time} or when \code{r_failures} events occur, whichever
#' comes first.
#'
#' @param n sample size.
#' @param dist a \code{\link{dist_spec}} object.
#' @param cen_time fixed time limit.
#' @param r_failures failure count limit.
#' @param seed optional integer seed.
#' @return A \code{censored_sample} object (S3 class) containing:
#'   \code{scheme} (character), \code{dist} (the \code{dist_spec}),
#'   \code{n} (integer sample size), \code{seed}, \code{t_true} (numeric
#'   vector of latent failure times), \code{x} (observed/censored values),
#'   \code{status} (integer: 0 = right-censored, 1 = event, 2 = left-censored,
#'   3 = interval-censored, 4 = middle-censored), and \code{time2}
#'   (upper interval bound where applicable, \code{NA} otherwise).
#' @export
#' @examples
#' d <- dist_spec(q = function(u, p) -log(1 - u) / p$rate,
#'                param = list(rate = 1), name = "Exp(1)")
#' r_hybrid_censor(30, d, cen_time = 2, r_failures = 15, seed = 1)
r_hybrid_censor <- function(n, dist, cen_time, r_failures, seed = NULL) {
  n <- .check_n(n)
  .validate_dist(dist)
  .apply_seed(seed)
  t_true <- .rng_stream(n, dist)
  stop <- .type2_stop(t_true, r_failures)
  stop_time <- min(cen_time, stop$r_th_time)
  event <- t_true <= stop_time
  x <- ifelse(event, t_true, stop_time)
  status <- ifelse(event, 1L, 0L)
  .as_censored_sample("hybrid", dist, n, seed, t_true, x, status,
                      extra = list(cen_time = cen_time,
                                   r_failures = min(as.integer(r_failures), n),
                                   stop_time = stop_time))
}

#' Hybrid Type-I censoring with progressive withdrawal
#'
#' At each time in \code{prog_times}, \code{scheme[j]} units are withdrawn;
#' remaining units are Type-I censored at \code{cen_time}.
#'
#' @param n sample size.
#' @param dist a \code{\link{dist_spec}} object.
#' @param cen_time final Type-I censoring time.
#' @param prog_times vector of progressive withdrawal times.
#' @param scheme removals at each progressive time.
#' @param seed optional integer seed.
#' @return A \code{censored_sample} object (S3 class) containing:
#'   \code{scheme} (character), \code{dist} (the \code{dist_spec}),
#'   \code{n} (integer sample size), \code{seed}, \code{t_true} (numeric
#'   vector of latent failure times), \code{x} (observed/censored values),
#'   \code{status} (integer: 0 = right-censored, 1 = event, 2 = left-censored,
#'   3 = interval-censored, 4 = middle-censored), and \code{time2}
#'   (upper interval bound where applicable, \code{NA} otherwise).
#' @export
#' @examples
#' d <- dist_spec(q = function(u, p) -log(1 - u) / p$rate,
#'                param = list(rate = 1), name = "Exp(1)")
#' r_hybrid_type1_censor(30, d, cen_time = 3,
#'                       prog_times = c(1, 2), scheme = c(2, 2), seed = 1)
r_hybrid_type1_censor <- function(n, dist, cen_time, prog_times, scheme,
                                  seed = NULL) {
  n <- .check_n(n)
  .validate_dist(dist)
  .apply_seed(seed)
  t_true <- .rng_stream(n, dist)
  prog_times <- sort(prog_times)
  scheme     <- as.integer(scheme)
  m          <- min(length(prog_times), length(scheme))
  x          <- t_true
  status     <- rep(1L, n)
  at_risk    <- rep(TRUE, n)

  for (j in seq_len(m)) {
    tau <- prog_times[j]
    at_risk[t_true <= tau & status == 1L] <- FALSE
    eligible <- which(at_risk & t_true > tau)
    n_rem <- min(scheme[j], length(eligible))
    if (n_rem > 0L) {
      rem <- sample(eligible, n_rem)
      x[rem] <- tau
      status[rem] <- 0L
      at_risk[rem] <- FALSE
    }
  }

  still <- which(at_risk)
  cens  <- t_true[still] > cen_time
  x[still[cens]] <- cen_time
  status[still[cens]] <- 0L

  .as_censored_sample("hybrid_type1", dist, n, seed, t_true, x, status,
                      extra = list(cen_time = cen_time,
                                   prog_times = prog_times,
                                   scheme = scheme))
}

#' Doubly Type-II censoring
#'
#' Classical doubly Type-II censoring: the \code{r1} smallest and \code{r2}
#' largest latent failure times are censored; the middle order statistics are
#' observed exactly.  Requires \code{r1 + r2 < n}.
#'
#' @param n sample size.
#' @param dist a \code{\link{dist_spec}} object.
#' @param r1 number of smallest order statistics left-censored.
#' @param r2 number of largest order statistics right-censored.
#' @param seed optional integer seed.
#' @return A \code{censored_sample} object (S3 class) containing:
#'   \code{scheme} (character), \code{dist} (the \code{dist_spec}),
#'   \code{n} (integer sample size), \code{seed}, \code{t_true} (numeric
#'   vector of latent failure times), \code{x} (observed/censored values),
#'   \code{status} (integer: 0 = right-censored, 1 = event, 2 = left-censored,
#'   3 = interval-censored, 4 = middle-censored), and \code{time2}
#'   (upper interval bound where applicable, \code{NA} otherwise).
#' @export
#' @examples
#' d <- dist_spec(q = function(u, p) -log(1 - u) / p$rate,
#'                param = list(rate = 1), name = "Exp(1)")
#' r_doubly_type2_censor(50, d, r1 = 5, r2 = 5, seed = 1)
r_doubly_type2_censor <- function(n, dist, r1, r2, seed = NULL) {
  n <- .check_n(n)
  .validate_dist(dist)
  .apply_seed(seed)
  r1 <- as.integer(r1)
  r2 <- as.integer(r2)
  if (r1 < 0L || r2 < 0L) {
    stop("'r1' and 'r2' must be non-negative.", call. = FALSE)
  }
  if (r1 + r2 >= n) {
    stop("'r1' + 'r2' must be less than 'n'.", call. = FALSE)
  }
  t_true <- .rng_stream(n, dist)
  ord <- order(t_true)
  r1 <- min(r1, n - 1L)
  r2 <- min(r2, n - r1 - 1L)

  x <- t_true
  status <- rep(1L, n)
  if (r1 > 0L) {
    left_c <- ord[seq_len(r1)]
    x[left_c] <- t_true[ord[r1 + 1L]]
    status[left_c] <- 2L
  }
  if (r2 > 0L) {
    right_c <- ord[seq(n - r2 + 1L, n)]
    x[right_c] <- t_true[ord[n - r2]]
    status[right_c] <- 0L
  }

  .as_censored_sample("doubly_type2", dist, n, seed, t_true, x, status,
                      extra = list(r1 = r1, r2 = r2))
}

#' Type-I hybrid censoring
#'
#' Experiment terminates at \code{cen_time} or when \code{r_failures} events
#' occur; if \code{r_failures} is reached first, a final Type-I censoring time
#' \code{cen_time} still applies to remaining units.
#'
#' @param n sample size.
#' @param dist a \code{\link{dist_spec}} object.
#' @param cen_time fixed censoring time.
#' @param r_failures failure count target.
#' @param seed optional integer seed.
#' @return A \code{censored_sample} object (S3 class) containing:
#'   \code{scheme} (character), \code{dist} (the \code{dist_spec}),
#'   \code{n} (integer sample size), \code{seed}, \code{t_true} (numeric
#'   vector of latent failure times), \code{x} (observed/censored values),
#'   \code{status} (integer: 0 = right-censored, 1 = event, 2 = left-censored,
#'   3 = interval-censored, 4 = middle-censored), and \code{time2}
#'   (upper interval bound where applicable, \code{NA} otherwise).
#' @export
#' @examples
#' d <- dist_spec(q = function(u, p) -log(1 - u) / p$rate,
#'                param = list(rate = 1), name = "Exp(1)")
#' r_type1_hybrid_censor(30, d, cen_time = 2, r_failures = 15, seed = 1)
r_type1_hybrid_censor <- function(n, dist, cen_time, r_failures,
                                  seed = NULL) {
  n <- .check_n(n)
  .validate_dist(dist)
  .apply_seed(seed)
  t_true <- .rng_stream(n, dist)
  r_failures <- min(as.integer(r_failures), n)
  ord  <- order(t_true)
  r_th <- t_true[ord[r_failures]]
  stop_time <- if (r_th <= cen_time) r_th else cen_time
  cens <- t_true > stop_time
  x    <- ifelse(cens, stop_time, t_true)
  status <- ifelse(cens, 0L, 1L)
  .as_censored_sample("type1_hybrid", dist, n, seed, t_true, x, status,
                      extra = list(cen_time = cen_time,
                                   r_failures = r_failures,
                                   stop_time = stop_time))
}

#' Hybrid Type-II censoring
#'
#' Progressive Type-II removals are applied, then the experiment terminates
#' at the \code{r_failures}-th remaining failure.
#'
#' @param n sample size.
#' @param dist a \code{\link{dist_spec}} object.
#' @param scheme progressive removal scheme before Type-II stop.
#' @param r_failures failure count for final Type-II termination.
#' @param seed optional integer seed.
#' @return A \code{censored_sample} object (S3 class) containing:
#'   \code{scheme} (character), \code{dist} (the \code{dist_spec}),
#'   \code{n} (integer sample size), \code{seed}, \code{t_true} (numeric
#'   vector of latent failure times), \code{x} (observed/censored values),
#'   \code{status} (integer: 0 = right-censored, 1 = event, 2 = left-censored,
#'   3 = interval-censored, 4 = middle-censored), and \code{time2}
#'   (upper interval bound where applicable, \code{NA} otherwise).
#' @export
#' @examples
#' d <- dist_spec(q = function(u, p) -log(1 - u) / p$rate,
#'                param = list(rate = 1), name = "Exp(1)")
#' r_hybrid_type2_censor(30, d, scheme = c(1, 1), r_failures = 10, seed = 1)
r_hybrid_type2_censor <- function(n, dist, scheme, r_failures,
                                  seed = NULL) {
  n <- .check_n(n)
  .validate_dist(dist)
  .apply_seed(seed)
  t_true  <- .rng_stream(n, dist)
  scheme  <- as.integer(scheme)
  pr <- .progressive_type2_removals(t_true, scheme)

  still <- which(!pr$removed)
  t_still <- t_true[still]
  r_failures <- min(as.integer(r_failures), length(still))
  stop <- .type2_stop(t_still, r_failures)
  r_th <- stop$r_th_time

  x      <- t_true
  status <- rep(1L, n)
  status[pr$removed] <- 0L
  x[pr$removed] <- pr$censor_at[pr$removed]
  cens <- !pr$removed & t_true > r_th
  x[cens] <- r_th
  status[cens] <- 0L

  .as_censored_sample("hybrid_type2", dist, n, seed, t_true, x, status,
                      extra = list(scheme = scheme,
                                   r_failures = r_failures,
                                   r_th_time = r_th))
}

# Type-II progressively hybrid censoring ---------------------------------------

#' Type-II progressively hybrid censoring
#'
#' Combines progressive Type-II censoring with a pre-set time limit
#' \code{cen_time}.  Progressive removals according to \code{scheme} are
#' applied at each failure.  The experiment terminates at
#' \eqn{\min(X_{m:m:n}, T)}, where \eqn{X_{m:m:n}} is the \code{m}-th
#' (i.e., \code{r_failures}-th) failure among the remaining units and
#' \eqn{T} is \code{cen_time}.  If fewer than \code{r_failures} events
#' occur before \code{cen_time}, all surviving units are censored at
#' \code{cen_time}.  See Kundu and Joarder (2006)
#' \doi{10.1016/j.csda.2005.05.002}.
#'
#' @param n sample size.
#' @param dist a \code{\link{dist_spec}} object.
#' @param scheme integer vector of removals at successive failure times.
#' @param r_failures target number of failures (default
#'   \eqn{\lfloor 0.7 n \rfloor}).
#' @param cen_time pre-set maximum experiment time.
#' @param seed optional integer seed.
#' @return A \code{censored_sample} object (S3 class) containing:
#'   \code{scheme} (character "progressive_hybrid_type2"), \code{dist} (the \code{dist_spec}),
#'   \code{n} (integer sample size), \code{seed}, \code{t_true} (numeric
#'   vector of latent failure times), \code{x} (observed/censored values),
#'   \code{status} (integer: 0 = right-censored, 1 = event, 2 = left-censored,
#'   3 = interval-censored, 4 = middle-censored), \code{time2}
#'   (upper interval bound where applicable, \code{NA} otherwise), and
#'   \code{extra} (list containing \code{scheme}, \code{r_failures}, \code{cen_time},
#'   and \code{stop_time}).
#' @export
#' @examples
#' d <- dist_spec(q = function(u, p) -log(1 - u) / p$rate,
#'                param = list(rate = 1), name = "Exp(1)")
#' r_progressive_hybrid_type2_censor(30, d, scheme = c(1, 1, 1),
#'                                   r_failures = 10, cen_time = 3, seed = 1)
r_progressive_hybrid_type2_censor <- function(n, dist, scheme, r_failures = NULL,
                                              cen_time, seed = NULL) {
  n <- .check_n(n)
  .validate_dist(dist)
  .apply_seed(seed)
  t_true  <- .rng_stream(n, dist)
  scheme  <- as.integer(scheme)
  if (is.null(r_failures)) r_failures <- .default_r_failures(n)
  r_failures <- min(as.integer(r_failures), n)

  # Progressive removals at ordered failure times
  ord     <- order(t_true)
  removed <- logical(n)
  censor_at <- rep(NA_real_, n)
  m <- min(length(scheme), n - 1L)

  n_observed <- 0L
  for (j in seq_len(n)) {
    fail_pos <- ord[j]
    if (removed[fail_pos]) next
    t_j <- t_true[fail_pos]

    # If this failure time exceeds cen_time, stop
    if (t_j > cen_time) break

    n_observed <- n_observed + 1L

    # Apply progressive removals if scheme still has entries
    if (n_observed <= m) {
      candidates <- which(!removed & t_true > t_j)
      # Exclude the current failure from candidates
      candidates <- setdiff(candidates, fail_pos)
      n_rem <- min(scheme[n_observed], length(candidates))
      if (n_rem > 0L) {
        to_remove <- if (n_rem == 1L) {
          candidates[sample.int(length(candidates), 1L)]
        } else {
          sample(candidates, n_rem)
        }
        removed[to_remove] <- TRUE
        censor_at[to_remove] <- t_j
      }
    }

    # Check if we reached the target number of failures
    if (n_observed >= r_failures) break
  }

  # Determine effective stop time
  # If r_failures observed, stop_time is the r-th failure time
  # Otherwise stop_time is cen_time
  if (n_observed >= r_failures) {
    # Find the r_failures-th observed failure time
    obs_failures <- ord[!removed[ord]]
    obs_times <- t_true[obs_failures]
    stop_time <- sort(obs_times[obs_times <= cen_time])[r_failures]
  } else {
    stop_time <- cen_time
  }

  # Build output
  x      <- t_true
  status <- rep(1L, n)

  # Units already removed progressively
  status[removed] <- 0L
  x[removed] <- censor_at[removed]

  # Units still at risk but failing after stop_time
  late <- !removed & t_true > stop_time
  x[late] <- stop_time
  status[late] <- 0L

  .as_censored_sample("progressive_hybrid_type2", dist, n, seed,
                      t_true, x, status,
                      extra = list(scheme = scheme,
                                   r_failures = r_failures,
                                   cen_time = cen_time,
                                   stop_time = stop_time))
}

make_exp_dist <- function(rate = 1) {
  dist_spec(
    d    = function(x, p) p$rate * exp(-p$rate * x),
    p    = function(x, p) 1 - exp(-p$rate * x),
    q    = function(u, p) -log(1 - u) / p$rate,
    surv = function(x, p) exp(-p$rate * x),
    lower = 0, upper = Inf,
    param = list(rate = rate),
    name  = "Exp(1)"
  )
}

# -- dist_spec -----------------------------------------------------------------

test_that("dist_spec validates input", {
  expect_error(dist_spec(), "Provide at least one")
  d <- make_exp_dist()
  expect_s3_class(d, "dist_spec")
})

test_that("dist_spec works with only PDF", {
  d <- dist_spec(d = function(x, p) p$rate * exp(-p$rate * x),
                 lower = 0, upper = Inf, param = list(rate = 1))
  expect_s3_class(d, "dist_spec")
})

# -- Reproducibility -----------------------------------------------------------

test_that("reproducibility: same seed gives same sample", {
  d <- make_exp_dist()
  a <- r_complete(100, d, seed = 123)
  b <- r_complete(100, d, seed = 123)
  expect_equal(a$t_true, b$t_true)
  c <- r_type1_censor(100, d, cen_time = 2, seed = 456)
  e <- r_type1_censor(100, d, cen_time = 2, seed = 456)
  expect_equal(c$x, e$x)
  expect_equal(c$status, e$status)
})

test_that("reproducibility: middle censoring with same seed", {
  d <- make_exp_dist()
  a <- r_middle_censor(100, d, lower = 0.5, upper = 1.5, seed = 77)
  b <- r_middle_censor(100, d, lower = 0.5, upper = 1.5, seed = 77)
  expect_equal(a$t_true, b$t_true)
  expect_equal(a$x, b$x)
  expect_equal(a$status, b$status)
  expect_equal(a$time2, b$time2)
})

# -- Complete sample -----------------------------------------------------------

test_that("complete sample has all events", {
  d <- make_exp_dist()
  s <- r_complete(50, d, seed = 1)
  expect_length(s$x, 50)
  expect_true(all(s$status == 1L))
  expect_equal(s$x, s$t_true)
})

# -- Type-II -------------------------------------------------------------------

test_that("type2 censoring yields at least r_failures events", {
  d <- make_exp_dist()
  s <- r_type2_censor(100, d, r_failures = 30, seed = 99)
  expect_gte(sum(s$status == 1L), 30L)
  if (any(s$status == 0L)) {
    expect_true(all(s$x[s$status == 0L] == s$r_th_time))
  }
})

# -- Doubly Type-II ------------------------------------------------------------

test_that("doubly type2 censors correct counts", {
  d <- make_exp_dist()
  s <- r_doubly_type2_censor(50, d, r1 = 5, r2 = 5, seed = 11)
  expect_equal(sum(s$status == 2L), 5L)
  expect_equal(sum(s$status == 0L), 5L)
  expect_equal(sum(s$status == 1L), 40L)
  expect_error(r_doubly_type2_censor(10, d, r1 = 5, r2 = 5, seed = 1),
               "r1.*r2")
})

# -- Middle censoring ----------------------------------------------------------

test_that("middle censoring status codes are correct", {
  d <- make_exp_dist()
  s <- r_middle_censor(200, d, lower = 0.5, upper = 1.5, seed = 42)
  # All status codes must be 1 (event) or 4 (middle-censored)
  expect_true(all(s$status %in% c(1L, 4L)))
  # With n = 200, some should be middle-censored
  expect_gt(sum(s$status == 4L), 0L)
  # Middle-censored: x stores the lower bound
  mc <- s$status == 4L
  expect_true(all(s$x[mc] == 0.5))
  # Middle-censored: time2 stores the upper bound
  expect_true(all(s$time2[mc] == 1.5))
  # Uncensored: x == t_true
  obs <- s$status == 1L
  expect_equal(s$x[obs], s$t_true[obs])
  # Middle-censored: t_true is in [lower, upper]
  expect_true(all(s$t_true[mc] >= 0.5 & s$t_true[mc] <= 1.5))
  # Uncensored: t_true is outside [lower, upper]
  expect_true(all(s$t_true[obs] < 0.5 | s$t_true[obs] > 1.5))
})

test_that("middle censoring validates lower < upper", {
  d <- make_exp_dist()
  expect_error(r_middle_censor(30, d, lower = 2, upper = 1, seed = 1),
               "lower.*upper")
  expect_error(r_middle_censor(30, d, lower = 1, upper = 1, seed = 1),
               "lower.*upper")
})

# -- Right / Left censoring ---------------------------------------------------

test_that("right censoring caps at cen_time", {
  d <- make_exp_dist()
  s <- r_right_censor(100, d, cen_time = 1, seed = 5)
  expect_true(all(s$x <= 1))
  expect_true(all(s$x[s$status == 0L] == 1))
  expect_equal(s$x[s$status == 1L], s$t_true[s$status == 1L])
})

test_that("left censoring floors at threshold", {
  d <- make_exp_dist()
  s <- r_left_censor(100, d, threshold = 0.5, seed = 5)
  expect_true(all(s$x >= 0.5 | s$status == 1L))
  expect_true(all(s$x[s$status == 2L] == 0.5))
})

# -- Progressive Type-II ------------------------------------------------------

test_that("progressive type2 removes units correctly", {
  d <- make_exp_dist()
  s <- r_progressive_type2_censor(50, d, scheme = c(2, 2, 2), seed = 10)
  expect_length(s$x, 50)
  expect_true(sum(s$status == 0L) <= 6L)  # at most sum(scheme) removed
  # All censored x values should be <= some failure time
  if (any(s$status == 0L)) {
    expect_true(all(s$x[s$status == 0L] <= max(s$t_true)))
  }
})

# -- Hybrid censoring ---------------------------------------------------------

test_that("hybrid censoring stops at min(cen_time, r-th failure)", {
  d <- make_exp_dist()
  s <- r_hybrid_censor(100, d, cen_time = 5, r_failures = 50, seed = 3)
  expect_length(s$x, 100)
  expect_true(all(s$x[s$status == 0L] == s$stop_time))
})

# -- Truncation ----------------------------------------------------------------

test_that("right truncation generates values <= right", {
  d <- make_exp_dist()
  s <- r_right_truncation(100, d, right = 2, seed = 10)
  expect_true(all(s$x <= 2))
  expect_true(all(s$status == 1L))
})

test_that("left truncation generates values >= left", {
  d <- make_exp_dist()
  s <- r_left_truncation(100, d, left = 0.5, seed = 10)
  expect_true(all(s$x >= 0.5))
  expect_true(all(s$status == 1L))
})

# -- Plot functions ------------------------------------------------------------

test_that("plot functions run without error", {
  d <- make_exp_dist()
  s <- r_random_censor(80, d, seed = 7)
  grDevices::pdf(NULL)
  on.exit(grDevices::dev.off(), add = TRUE)
  expect_invisible(plot_censor_hist(s))
  expect_invisible(plot_censor_dot(s))
  expect_invisible(plot_censor_acf(s))
  expect_invisible(plot_censor_diagnostics(s))
})

test_that("plot functions work for middle censoring", {
  d <- make_exp_dist()
  s <- r_middle_censor(80, d, lower = 0.5, upper = 1.5, seed = 7)
  grDevices::pdf(NULL)
  on.exit(grDevices::dev.off(), add = TRUE)
  expect_invisible(plot_censor_hist(s))
  expect_invisible(plot_censor_dot(s))
  expect_invisible(plot_censor_acf(s))
})

# -- All scheme functions return censored_sample -------------------------------

test_that("all scheme functions return censored_sample", {
  d <- make_exp_dist()
  n <- 40L
  schemes <- list(
    r_complete(n, d, seed = 1),
    r_right_truncation(n, d, right = 3, seed = 1),
    r_left_truncation(n, d, left = 0.5, seed = 2),
    r_random_censor(n, d, seed = 3),
    r_right_censor(n, d, cen_time = 2, seed = 4),
    r_left_censor(n, d, threshold = 0.3, seed = 5),
    r_interval_censor(n, d, width = 0.5, seed = 6),
    r_middle_censor(n, d, lower = 0.5, upper = 1.5, seed = 6),
    r_block_random_censor(n, d, seed = 7),
    r_type1_censor(n, d, cen_time = 2, seed = 8),
    r_type2_censor(n, d, r_failures = 20, seed = 9),
    r_progressive_type2_censor(n, d, seed = 10),
    r_progressive_first_failure(n, d, seed = 11),
    r_hybrid_censor(n, d, cen_time = 2, r_failures = 15, seed = 12),
    r_hybrid_type1_censor(n, d, cen_time = 3,
                          prog_times = c(1, 2), scheme = c(2, 2), seed = 13),
    r_doubly_type2_censor(n, d, r1 = 5, r2 = 5, seed = 14),
    r_type1_hybrid_censor(n, d, cen_time = 2, r_failures = 15, seed = 15),
    r_hybrid_type2_censor(n, d, scheme = c(1, 1), r_failures = 10, seed = 16),
    r_progressive_hybrid_type2_censor(n, d, scheme = c(1, 1), r_failures = 10, cen_time = 3, seed = 17)
  )
  for (s in schemes) {
    expect_s3_class(s, "censored_sample")
    expect_equal(length(s$x), n)
    expect_equal(length(s$status), n)
    expect_equal(length(s$t_true), n)
  }
  j1 <- r_joint_type1_censor(20, d, cen_time = 2, k = 2, seed = 17)
  j2 <- r_joint_type2_censor(20, d, r_failures = 15, k = 2, seed = 18)
  bj <- r_bjpt2_censor(20, d, k = 2, seed = 19)
  expect_s3_class(j1, "censored_sample_list")
  expect_s3_class(j2, "censored_sample_list")
  expect_s3_class(bj, "censored_sample_list")
})

# -- Type-II Progressively Hybrid Censoring ------------------------------------

test_that("Type-II progressively hybrid censoring stops at min(m-th failure, T)", {
  d <- make_exp_dist()
  # Case I: m-th failure before T (T = 5, m = 10)
  s1 <- r_progressive_hybrid_type2_censor(50, d, scheme = rep(1L, 10), r_failures = 10, cen_time = 5, seed = 42)
  expect_s3_class(s1, "censored_sample")
  expect_length(s1$x, 50)
  # Check that stop_time is the 10th failure time
  expect_true(s1$stop_time < 5)
  expect_equal(sum(s1$status == 1L), 10L)
  
  # Case II: m-th failure after T (T = 0.5, m = 40)
  s2 <- r_progressive_hybrid_type2_censor(50, d, scheme = rep(1L, 40), r_failures = 40, cen_time = 0.5, seed = 42)
  expect_equal(s2$stop_time, 0.5)
  expect_true(sum(s2$status == 1L) < 40L)
  expect_true(all(s2$x[s2$status == 0L] <= 0.5))
})

# -- Coercion methods ----------------------------------------------------------

test_that("as.vector and as.data.frame work", {
  d <- make_exp_dist()
  s <- r_complete(20, d, seed = 1)
  v <- as.vector(s)
  expect_length(v, 20)
  df <- as.data.frame(s)
  expect_s3_class(df, "data.frame")
  expect_equal(nrow(df), 20)
  expect_true(all(c("x", "status", "t_true", "time2") %in% names(df)))
})

# -- Edge cases ----------------------------------------------------------------

test_that("n = 1 works for all basic schemes", {
  d <- make_exp_dist()
  expect_s3_class(r_complete(1, d, seed = 1), "censored_sample")
  expect_s3_class(r_right_censor(1, d, cen_time = 2, seed = 1), "censored_sample")
  expect_s3_class(r_type1_censor(1, d, cen_time = 2, seed = 1), "censored_sample")
  expect_s3_class(r_type2_censor(1, d, r_failures = 1, seed = 1), "censored_sample")
  expect_s3_class(r_middle_censor(1, d, lower = 0.5, upper = 1.5, seed = 1),
                  "censored_sample")
  expect_s3_class(r_progressive_hybrid_type2_censor(1, d, scheme = 0L, r_failures = 1, cen_time = 2, seed = 1),
                  "censored_sample")
})

test_that("invalid n is rejected", {
  d <- make_exp_dist()
  expect_error(r_complete(0, d), "positive")
  expect_error(r_complete(-5, d), "positive")
})

# Internal utilities for UniCensor -------------------------------------------

#' @keywords internal
.validate_dist <- function(dist) {
  if (!inherits(dist, "dist_spec")) {
    stop("'dist' must be a dist_spec object; use dist_spec() to create one.",
         call. = FALSE)
  }
  invisible(dist)
}

#' @keywords internal
.apply_seed <- function(seed) {
  if (!is.null(seed)) {
    if (!is.numeric(seed) || length(seed) != 1L || is.na(seed)) {
      stop("'seed' must be a single numeric value or NULL.", call. = FALSE)
    }
    set.seed(as.integer(seed))
  }
  invisible(seed)
}

#' @keywords internal
.as_censored_sample <- function(scheme, dist, n, seed, t_true, x, status,
                                time2 = rep(NA_real_, length(x)),
                                extra = list()) {
  out <- list(
    scheme  = scheme,
    dist    = dist,
    n       = as.integer(n),
    seed    = seed,
    t_true  = as.numeric(t_true),
    x       = as.numeric(x),
    status  = as.integer(status),
    time2   = as.numeric(time2)
  )
  if (length(extra)) out <- c(out, extra)
  class(out) <- "censored_sample"
  out
}

#' @keywords internal
.rng_stream <- function(n, dist) {
  if (!is.null(dist$q)) {
    u <- stats::runif(n)
    return(vapply(u, function(ui) dist$q(ui, dist$param), numeric(1)))
  }
  u <- stats::runif(n)
  vapply(seq_len(n), function(i) {
    .invert_survival(u[i], dist)
  }, numeric(1))
}

#' @keywords internal
.invert_survival <- function(u, dist, interval = NULL) {
  if (is.null(interval)) {
    interval <- c(dist$lower, dist$upper)
    if (!is.finite(interval[2])) interval[2] <- dist$lower + 1e4
  }
  target <- 1 - u
  if (!is.null(dist$surv)) {
    f <- function(t) dist$surv(t, dist$param) - target
  } else if (!is.null(dist$p)) {
    f <- function(t) dist$p(t, dist$param) - u
  } else if (!is.null(dist$d)) {
    ll <- dist$lower
    dens <- function(x) dist$d(x, dist$param)
    cdf_at <- function(t) {
      if (!is.finite(t) || t <= ll) return(0)
      if (is.finite(dist$upper) && t >= dist$upper) return(1)
      stats::integrate(dens, ll, t, rel.tol = 1e-8, subdivisions = 200L)$value
    }
    f <- function(t) cdf_at(t) - u
  } else {
    stop("dist_spec must provide at least one of q, surv, p, or d.",
         call. = FALSE)
  }
  res <- suppressWarnings(tryCatch(
    stats::uniroot(f, interval = interval, tol = 1e-8, extendInt = "upX"),
    error = function(e) NULL
  ))
  if (is.null(res)) {
    stop("Could not invert the distribution on support [",
         interval[1], ", ", interval[2], "].", call. = FALSE)
  }
  res$root
}

#' @keywords internal
.rng_truncated <- function(n, dist, left = dist$lower, right = dist$upper) {
  if (!is.null(dist$q) && !is.null(dist$p)) {
    pl <- dist$p(left, dist$param)
    pr <- dist$p(right, dist$param)
    u  <- stats::runif(n, min = pl, max = pr)
    return(vapply(u, function(ui) dist$q(ui, dist$param), numeric(1)))
  }
  vapply(seq_len(n), function(i) {
    u <- stats::runif(1)
    if (!is.null(dist$surv)) {
      sl <- dist$surv(left, dist$param)
      sr <- dist$surv(right, dist$param)
      target <- sr + u * (sl - sr)
      f <- function(t) dist$surv(t, dist$param) - target
    } else {
      pl <- dist$p(left, dist$param)
      pr <- dist$p(right, dist$param)
      target <- pl + u * (pr - pl)
      f <- function(t) dist$p(t, dist$param) - target
    }
    stats::uniroot(f, interval = c(left, right), tol = 1e-8)$root
  }, numeric(1))
}

#' @keywords internal
.check_n <- function(n) {
  n <- as.integer(n)
  if (length(n) != 1L || is.na(n) || n <= 0L) {
    stop("'n' must be a positive integer.", call. = FALSE)
  }
  n
}

#' @keywords internal
.default_cen_time <- function(t_true, prob = 0.7) {
  unname(stats::quantile(t_true, prob, names = FALSE))
}

#' @keywords internal
.default_r_failures <- function(n, prop = 0.7) {
  max(1L, floor(prop * n))
}

#' @keywords internal
.type2_stop <- function(t_true, r_failures) {
  r_failures <- min(as.integer(r_failures), length(t_true))
  ord  <- order(t_true)
  r_th <- t_true[ord[r_failures]]
  event <- t_true <= r_th
  list(
    x = ifelse(event, t_true, r_th),
    status = ifelse(event, 1L, 0L),
    r_th_time = r_th
  )
}

#' @keywords internal
.progressive_type2_removals <- function(t_true, scheme) {
  n <- length(t_true)
  scheme <- as.integer(scheme)
  ord <- order(t_true)
  removed <- logical(n)
  censor_at <- rep(NA_real_, n)
  m <- min(length(scheme), n - 1L)

  for (j in seq_len(m)) {
    fail_pos <- ord[j]
    if (removed[fail_pos]) next          # already removed; skip
    t_j <- t_true[fail_pos]
    # Candidates: not yet removed AND with failure time > t_j
    candidates <- which(!removed & t_true > t_j)
    n_rem <- min(scheme[j], length(candidates))
    if (n_rem > 0L) {
      to_remove <- if (n_rem == 1L) candidates[sample.int(length(candidates), 1L)] else sample(candidates, n_rem)
      removed[to_remove] <- TRUE
      censor_at[to_remove] <- t_j
    }
  }

  list(removed = removed, censor_at = censor_at)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/censor_schemes.R
\name{r_complete}
\alias{r_complete}
\title{Generate a complete (uncensored) random sample}
\usage{
r_complete(n, dist, seed = NULL)
}
\arguments{
\item{n}{sample size.}

\item{dist}{a \code{\link{dist_spec}} object.}

\item{seed}{optional integer seed for reproducibility.}
}
\value{
A \code{censored_sample} object (S3 class) containing:
\code{scheme} (character), \code{dist} (the \code{dist_spec}),
\code{n} (integer sample size), \code{seed}, \code{t_true} (numeric
vector of latent failure times), \code{x} (observed/censored values),
\code{status} (integer: 1 = event for all observations in a complete
sample), and \code{time2} (\code{NA} for all).
}
\description{
Generate a complete (uncensored) random sample
}
\examples{
exp_dist <- dist_spec(
  q = function(u, p) -log(1 - u) / p$rate,
  surv = function(x, p) exp(-p$rate * x),
  param = list(rate = 1), name = "Exp(1)"
)
samp <- r_complete(50, exp_dist, seed = 42)
table(samp$status)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/censor_schemes.R
\name{r_progressive_hybrid_type2_censor}
\alias{r_progressive_hybrid_type2_censor}
\title{Type-II progressively hybrid censoring}
\usage{
r_progressive_hybrid_type2_censor(
  n,
  dist,
  scheme,
  r_failures = NULL,
  cen_time,
  seed = NULL
)
}
\arguments{
\item{n}{sample size.}

\item{dist}{a \code{\link{dist_spec}} object.}

\item{scheme}{integer vector of removals at successive failure times.}

\item{r_failures}{target number of failures (default
\eqn{\lfloor 0.7 n \rfloor}).}

\item{cen_time}{pre-set maximum experiment time.}

\item{seed}{optional integer seed.}
}
\value{
A \code{censored_sample} object (S3 class) containing:
\code{scheme} (character "progressive_hybrid_type2"), \code{dist} (the \code{dist_spec}),
\code{n} (integer sample size), \code{seed}, \code{t_true} (numeric
vector of latent failure times), \code{x} (observed/censored values),
\code{status} (integer: 0 = right-censored, 1 = event, 2 = left-censored,
3 = interval-censored, 4 = middle-censored), \code{time2}
(upper interval bound where applicable, \code{NA} otherwise), and
\code{extra} (list containing \code{scheme}, \code{r_failures}, \code{cen_time},
and \code{stop_time}).
}
\description{
Combines progressive Type-II censoring with a pre-set time limit
\code{cen_time}.  Progressive removals according to \code{scheme} are
applied at each failure.  The experiment terminates at
\eqn{\min(X_{m:m:n}, T)}, where \eqn{X_{m:m:n}} is the \code{m}-th
(i.e., \code{r_failures}-th) failure among the remaining units and
\eqn{T} is \code{cen_time}.  If fewer than \code{r_failures} events
occur before \code{cen_time}, all surviving units are censored at
\code{cen_time}.  See Kundu and Joarder (2006)
\doi{10.1016/j.csda.2005.05.002}.
}
\examples{
d <- dist_spec(q = function(u, p) -log(1 - u) / p$rate,
               param = list(rate = 1), name = "Exp(1)")
r_progressive_hybrid_type2_censor(30, d, scheme = c(1, 1, 1),
                                  r_failures = 10, cen_time = 3, seed = 1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_spec.R
\name{print.censored_sample}
\alias{print.censored_sample}
\title{Print method for censored_sample}
\usage{
\method{print}{censored_sample}(x, ...)
}
\arguments{
\item{x}{a \code{censored_sample} object.}

\item{...}{ignored.}
}
\value{
Invisibly, the input \code{censored_sample} object \code{x}.
Called for its side effect of printing the censoring scheme, distribution
name, sample size, seed, and a frequency table of status codes.
}
\description{
Print method for censored_sample
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_spec.R
\name{dist_spec}
\alias{dist_spec}
\title{Define a univariate distribution for simulation}
\usage{
dist_spec(
  d = NULL,
  p = NULL,
  q = NULL,
  surv = NULL,
  lower = 0,
  upper = Inf,
  param = list(),
  name = "custom"
)
}
\arguments{
\item{d}{function; PDF \eqn{f(x, \theta)}.}

\item{p}{function; CDF \eqn{F(x, \theta)}.}

\item{q}{function; quantile function \eqn{F^{-1}(u, \theta)}.}

\item{surv}{function; survival function \eqn{S(x, \theta)}.}

\item{lower}{numeric; lower support bound (default \code{0}).}

\item{upper}{numeric; upper support bound (default \code{Inf}).}

\item{param}{named list of distribution parameters (default empty).}

\item{name}{character label for plots (default \code{"custom"}).}
}
\value{
An object of class \code{"dist_spec"}, which is a list with
elements \code{d}, \code{p}, \code{q}, \code{surv} (user-supplied
functions or \code{NULL}), \code{lower} and \code{upper} (support
bounds), \code{param} (named list of parameters), and \code{name}
(character label).
}
\description{
Creates a reusable distribution specification passed to all
\pkg{UniCensor} sampling functions.  At least one of \code{q},
\code{surv}, \code{p}, or \code{d} must be supplied for random
number generation via inverse transform sampling.
}
\examples{
# Exponential(rate = 1)
exp_dist <- dist_spec(
  d    = function(x, p) p$rate * exp(-p$rate * x),
  p    = function(x, p) 1 - exp(-p$rate * x),
  q    = function(u, p) -log(1 - u) / p$rate,
  surv = function(x, p) exp(-p$rate * x),
  lower = 0, upper = Inf,
  param = list(rate = 1),
  name  = "Exp(1)"
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_spec.R
\name{as.data.frame.censored_sample}
\alias{as.data.frame.censored_sample}
\title{Coerce censored sample to data frame}
\usage{
\method{as.data.frame}{censored_sample}(x, row.names = NULL, optional = FALSE, ...)
}
\arguments{
\item{x}{a \code{censored_sample} object.}

\item{row.names}{ignored.}

\item{optional}{ignored.}

\item{...}{ignored.}
}
\value{
A data frame with columns \code{x} (observed values),
\code{status} (integer status code: 0 = right-censored, 1 = event,
2 = left-censored, 3 = interval-censored, 4 = middle-censored),
\code{t_true} (latent true failure times), and \code{time2} (upper
interval bound for interval- or middle-censored observations,
\code{NA} otherwise).
}
\description{
Coerce censored sample to data frame
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots.R
\name{plot_censor_diagnostics}
\alias{plot_censor_diagnostics}
\title{Combined diagnostic panel: histogram, dot plot, and ACF}
\usage{
plot_censor_diagnostics(samp, dist = NULL, use_latent = TRUE, ...)
}
\arguments{
\item{samp}{a \code{censored_sample} object.}

\item{dist}{optional \code{dist_spec}.}

\item{use_latent}{passed to individual plot functions.}

\item{...}{ignored.}
}
\value{
Invisibly, the input \code{censored_sample} object \code{samp}.
Called primarily for its side effect of producing a combined diagnostic
plot panel.
}
\description{
Produces a three-panel plot (histogram with PDF overlay, dot plot, and
ACF) for a censored sample.  Graphical parameters are saved before
modification and restored on exit via \code{\link{on.exit}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots.R
\name{plot_censor_dot}
\alias{plot_censor_dot}
\title{Dot plot (index plot) of sample values}
\usage{
plot_censor_dot(samp, use_latent = TRUE, pch = NULL, ...)
}
\arguments{
\item{samp}{a \code{censored_sample} object.}

\item{use_latent}{logical; plot latent (\code{t_true}) or observed
(\code{x}) values.}

\item{pch}{plotting character; censored points use \code{pch = 1},
events use \code{pch = 16}.}

\item{...}{passed to \code{\link[graphics]{plot}}.}
}
\value{
Invisibly, a list with elements \code{index} (integer vector of
observation indices) and \code{values} (numeric vector of plotted
values).  Called primarily for its side effect of producing a dot
(index) plot distinguishing events from censored observations.
}
\description{
Dot plot (index plot) of sample values
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/censor_schemes.R
\name{r_type2_censor}
\alias{r_type2_censor}
\title{Type-II (failure-terminated) censoring}
\usage{
r_type2_censor(n, dist, r_failures = NULL, seed = NULL)
}
\arguments{
\item{n}{sample size.}

\item{dist}{a \code{\link{dist_spec}} object.}

\item{r_failures}{number of failures before termination (default
\eqn{\lfloor 0.7n \rfloor}).}

\item{seed}{optional integer seed.}
}
\value{
A \code{censored_sample} object (S3 class) containing:
\code{scheme} (character), \code{dist} (the \code{dist_spec}),
\code{n} (integer sample size), \code{seed}, \code{t_true} (numeric
vector of latent failure times), \code{x} (observed/censored values),
\code{status} (integer: 0 = right-censored, 1 = event, 2 = left-censored,
3 = interval-censored, 4 = middle-censored), and \code{time2}
(upper interval bound where applicable, \code{NA} otherwise).
}
\description{
Experiment stops at the \code{r_failures}-th observed failure; remaining
units are right-censored at that time.
}
\examples{
d <- dist_spec(q = function(u, p) -log(1 - u) / p$rate,
               param = list(rate = 1), name = "Exp(1)")
r_type2_censor(30, d, r_failures = 20, seed = 1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_spec.R
\name{as.vector.censored_sample}
\alias{as.vector.censored_sample}
\title{Extract observed values from a censored sample}
\usage{
\method{as.vector}{censored_sample}(x, ...)
}
\arguments{
\item{x}{a \code{censored_sample} object.}

\item{...}{ignored.}
}
\value{
Numeric vector of observed values (failures and censoring times).
}
\description{
Extract observed values from a censored sample
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_spec.R
\name{print.dist_spec}
\alias{print.dist_spec}
\title{Print method for dist_spec}
\usage{
\method{print}{dist_spec}(x, ...)
}
\arguments{
\item{x}{a \code{dist_spec} object.}

\item{...}{ignored.}
}
\value{
Invisibly, the input \code{dist_spec} object \code{x}.  Called
for its side effect of printing a human-readable summary of the
distribution name, support, parameters, and available functions.
}
\description{
Print method for dist_spec
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots.R
\name{plot_censor_hist}
\alias{plot_censor_hist}
\title{Histogram of observed values with true density overlay}
\usage{
plot_censor_hist(samp, dist = NULL, use_latent = TRUE, nbreaks = 30L, ...)
}
\arguments{
\item{samp}{a \code{censored_sample} object.}

\item{dist}{optional \code{dist_spec}; defaults to \code{samp$dist}.}

\item{use_latent}{logical; plot histogram of latent failures
(\code{t_true}) instead of observed \code{x} (default \code{TRUE} for
verifying the underlying distribution).}

\item{nbreaks}{number of histogram breaks (default \code{30}).}

\item{...}{graphical parameters passed to \code{\link[graphics]{hist}}.}
}
\value{
Invisibly, an object of class \code{"histogram"} as returned by
\code{\link[graphics]{hist}}, containing break points, counts, and
density values.  Called primarily for its side effect of producing a
histogram plot with an optional PDF overlay.
}
\description{
For censored samples, uses latent \code{t_true} values to overlay the
target density when available; observed \code{x} values are histogrammed
with a note on censoring status.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/censor_schemes.R
\name{r_joint_type1_censor}
\alias{r_joint_type1_censor}
\title{Joint Type-I censoring across multiple samples}
\usage{
r_joint_type1_censor(n, dist, cen_time, k = 2L, seed = NULL)
}
\arguments{
\item{n}{sample size per group.}

\item{dist}{a \code{\link{dist_spec}} object.}

\item{cen_time}{fixed censoring time.}

\item{k}{number of independent samples (default \code{2}).}

\item{seed}{optional integer seed.}
}
\value{
A list of \code{censored_sample} objects (class
\code{"censored_sample_list"}), one per group.  Each element contains
\code{scheme}, \code{dist}, \code{n}, \code{seed}, \code{t_true},
\code{x}, \code{status} (0 = right-censored, 1 = event), \code{time2},
and extra fields \code{group} and \code{k}.
}
\description{
Generates \code{k} independent samples of size \code{n} each, all
censored at a common fixed time \code{cen_time}.
}
\examples{
d <- dist_spec(q = function(u, p) -log(1 - u) / p$rate,
               param = list(rate = 1), name = "Exp(1)")
r_joint_type1_censor(20, d, cen_time = 2, seed = 1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/UniCensor-package.R
\docType{package}
\name{UniCensor-package}
\alias{UniCensor}
\alias{UniCensor-package}
\title{UniCensor: Reproducible Univariate Samples Under Censoring Schemes}
\description{
The \pkg{UniCensor} package simulates censored or truncated random samples
from any user-defined univariate distribution.  Supply the PDF, CDF,
survival function, support bounds, parameters, sample size, and a seed;
identical inputs always produce identical output.
}
\section{Distribution specification}{

Use \code{\link{dist_spec}} to define a distribution object, then pass it
to any scheme-specific generator (e.g. \code{\link{r_type1_censor}}).
}

\section{Censoring schemes (one function each)}{

\describe{
\item{\code{\link{r_complete}}}{Uncensored (complete) data.}
\item{\code{\link{r_right_truncation}}}{Right truncation.}
\item{\code{\link{r_left_truncation}}}{Left truncation.}
\item{\code{\link{r_random_censor}}}{Random censoring.}
\item{\code{\link{r_right_censor}}}{Right censoring.}
\item{\code{\link{r_left_censor}}}{Left censoring.}
\item{\code{\link{r_interval_censor}}}{Interval censoring.}
\item{\code{\link{r_middle_censor}}}{Middle censoring.}
\item{\code{\link{r_block_random_censor}}}{Block random censoring.}
\item{\code{\link{r_bjpt2_censor}}}{Balanced joint progressive Type-II.}
\item{\code{\link{r_progressive_first_failure}}}{Progressive first failure.}
\item{\code{\link{r_joint_type1_censor}}}{Joint Type-I censoring.}
\item{\code{\link{r_type1_censor}}}{Type-I (time-terminated).}
\item{\code{\link{r_type2_censor}}}{Type-II (failure-terminated).}
\item{\code{\link{r_progressive_type2_censor}}}{Progressive Type-II.}
\item{\code{\link{r_progressive_hybrid_type2_censor}}}{Type-II progressively hybrid.}
\item{\code{\link{r_joint_type2_censor}}}{Joint Type-II censoring.}
\item{\code{\link{r_hybrid_censor}}}{Hybrid censoring.}
\item{\code{\link{r_hybrid_type1_censor}}}{Hybrid Type-I censoring.}
\item{\code{\link{r_doubly_type2_censor}}}{Doubly Type-II censoring.}
\item{\code{\link{r_type1_hybrid_censor}}}{Type-I hybrid censoring.}
\item{\code{\link{r_hybrid_type2_censor}}}{Hybrid Type-II censoring.}
}
}

\section{Diagnostic plots}{

\code{\link{plot_censor_hist}}, \code{\link{plot_censor_dot}},
\code{\link{plot_censor_acf}}, and \code{\link{plot_censor_diagnostics}}
(all three panels).
}

\author{
Shikhar Tyagi \email{shikhar1093tyagi@gmail.com}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots.R
\name{plot_censor_acf}
\alias{plot_censor_acf}
\title{Autocorrelation function plot}
\usage{
plot_censor_acf(samp, use_latent = TRUE, max_lag = NULL, ...)
}
\arguments{
\item{samp}{a \code{censored_sample} object.}

\item{use_latent}{logical; compute ACF on \code{t_true} (default) or
observed \code{x}.}

\item{max_lag}{maximum lag (default \code{min(20, n/4)}).}

\item{...}{passed to \code{\link[graphics]{plot}}.}
}
\value{
Invisibly, an object of class \code{"acf"} as returned by
\code{\link[stats]{acf}}, containing the autocorrelation values at each
lag.  Called primarily for its side effect of producing an ACF plot
with 95\\% confidence bands.
}
\description{
Plots the ACF of latent failure times to assess independence of the
underlying RNG stream (should resemble white noise for valid simulation).
}
